@@ -1,11 +1,12 @@
 use std::borrow::Cow;
+use std::ffi::CString;
 
+use crate::utils::{Ctx, LitCStr};
 use crate::{
     attributes::{self, get_pyo3_options, take_attributes, NameAttribute},
     deprecations::Deprecations,
 };
-use proc_macro2::{Ident, TokenStream};
-use quote::quote;
+use proc_macro2::{Ident, Span};
 use syn::{
     ext::IdentExt,
     parse::{Parse, ParseStream},
@@ -13,12 +14,12 @@ use syn::{
     Result,
 };
 
-pub struct ConstSpec {
+pub struct ConstSpec<'ctx> {
     pub rust_ident: syn::Ident,
-    pub attributes: ConstAttributes,
+    pub attributes: ConstAttributes<'ctx>,
 }
 
-impl ConstSpec {
+impl ConstSpec<'_> {
     pub fn python_name(&self) -> Cow<'_, Ident> {
         if let Some(name) = &self.attributes.name {
             Cow::Borrowed(&name.value.0)
@@ -28,16 +29,16 @@ impl ConstSpec {
     }
 
     /// Null-terminated Python name
-    pub fn null_terminated_python_name(&self) -> TokenStream {
-        let name = format!("{}\0", self.python_name());
-        quote!({#name})
+    pub fn null_terminated_python_name(&self, ctx: &Ctx) -> LitCStr {
+        let name = self.python_name().to_string();
+        LitCStr::new(CString::new(name).unwrap(), Span::call_site(), ctx)
     }
 }
 
-pub struct ConstAttributes {
+pub struct ConstAttributes<'ctx> {
     pub is_class_attr: bool,
     pub name: Option<NameAttribute>,
-    pub deprecations: Deprecations,
+    pub deprecations: Deprecations<'ctx>,
 }
 
 pub enum PyO3ConstAttribute {
@@ -55,12 +56,12 @@ impl Parse for PyO3ConstAttribute {
     }
 }
 
-impl ConstAttributes {
-    pub fn from_attrs(attrs: &mut Vec<syn::Attribute>) -> syn::Result<Self> {
+impl<'ctx> ConstAttributes<'ctx> {
+    pub fn from_attrs(attrs: &mut Vec<syn::Attribute>, ctx: &'ctx Ctx) -> syn::Result<Self> {
         let mut attributes = ConstAttributes {
             is_class_attr: false,
             name: None,
-            deprecations: Deprecations::new(),
+            deprecations: Deprecations::new(ctx),
         };
 
         take_attributes(attrs, |attr| {
