@@ -1,5 +1,3 @@
-use std::cmp::max;
-
 use proc_macro2::{Span, TokenStream};
 use quote::ToTokens;
 use syn::{
@@ -12,13 +10,10 @@ use syn::{
 
 use crate::{
     attributes::{kw, KeywordAttribute},
-    deprecations::{Deprecation, Deprecations},
-    method::FnArg,
-    pyfunction::Argument,
+    method::{FnArg, RegularArg},
 };
 
-use super::DeprecatedArgs;
-
+#[derive(Clone)]
 pub struct Signature {
     paren_token: syn::token::Paren,
     pub items: Punctuated<SignatureItem, Token![,]>,
@@ -27,10 +22,11 @@ pub struct Signature {
 impl Parse for Signature {
     fn parse(input: ParseStream<'_>) -> syn::Result<Self> {
         let content;
-        Ok(Signature {
-            paren_token: syn::parenthesized!(content in input),
-            items: content.parse_terminated(SignatureItem::parse)?,
-        })
+        let paren_token = syn::parenthesized!(content in input);
+
+        let items = content.parse_terminated(SignatureItem::parse, Token![,])?;
+
+        Ok(Signature { paren_token, items })
     }
 }
 
@@ -41,35 +37,35 @@ impl ToTokens for Signature {
     }
 }
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 pub struct SignatureItemArgument {
     pub ident: syn::Ident,
     pub eq_and_default: Option<(Token![=], syn::Expr)>,
 }
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 pub struct SignatureItemPosargsSep {
     pub slash: Token![/],
 }
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 pub struct SignatureItemVarargsSep {
     pub asterisk: Token![*],
 }
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 pub struct SignatureItemVarargs {
     pub sep: SignatureItemVarargsSep,
     pub ident: syn::Ident,
 }
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 pub struct SignatureItemKwargs {
     pub asterisks: (Token![*], Token![*]),
     pub ident: syn::Ident,
 }
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 pub enum SignatureItem {
     Argument(Box<SignatureItemArgument>),
     PosargsSep(SignatureItemPosargsSep),
@@ -200,6 +196,16 @@ impl ToTokens for SignatureItemPosargsSep {
 }
 
 pub type SignatureAttribute = KeywordAttribute<kw::signature, Signature>;
+pub type ConstructorAttribute = KeywordAttribute<kw::constructor, Signature>;
+
+impl ConstructorAttribute {
+    pub fn into_signature(self) -> SignatureAttribute {
+        SignatureAttribute {
+            kw: kw::signature(self.kw.span),
+            value: self.value,
+        }
+    }
+}
 
 #[derive(Default)]
 pub struct PythonSignature {
@@ -356,26 +362,39 @@ impl<'a> FunctionSignature<'a> {
 
         let mut next_non_py_argument_checked = |name: &syn::Ident| {
             for fn_arg in args_iter.by_ref() {
-                if fn_arg.py {
-                    // If the user incorrectly tried to include py: Python in the
-                    // signature, give a useful error as a hint.
-                    ensure_spanned!(
-                        name != fn_arg.name,
-                        name.span() => "arguments of type `Python` must not be part of the signature"
-                    );
-                    // Otherwise try next argument.
-                    continue;
+                match fn_arg {
+                    crate::method::FnArg::Py(..) => {
+                        // If the user incorrectly tried to include py: Python in the
+                        // signature, give a useful error as a hint.
+                        ensure_spanned!(
+                            name != fn_arg.name(),
+                            name.span() => "arguments of type `Python` must not be part of the signature"
+                        );
+                        // Otherwise try next argument.
+                        continue;
+                    }
+                    crate::method::FnArg::CancelHandle(..) => {
+                        // If the user incorrectly tried to include cancel: CoroutineCancel in the
+                        // signature, give a useful error as a hint.
+                        ensure_spanned!(
+                            name != fn_arg.name(),
+                            name.span() => "`cancel_handle` argument must not be part of the signature"
+                        );
+                        // Otherwise try next argument.
+                        continue;
+                    }
+                    _ => {
+                        ensure_spanned!(
+                            name == fn_arg.name(),
+                            name.span() => format!(
+                                "expected argument from function definition `{}` but got argument `{}`",
+                                fn_arg.name().unraw(),
+                                name.unraw(),
+                            )
+                        );
+                        return Ok(fn_arg);
+                    }
                 }
-
-                ensure_spanned!(
-                    name == fn_arg.name,
-                    name.span() => format!(
-                        "expected argument from function definition `{}` but got argument `{}`",
-                        fn_arg.name.unraw(),
-                        name.unraw(),
-                    )
-                );
-                return Ok(fn_arg);
             }
             bail_spanned!(
                 name.span() => "signature entry does not have a corresponding function argument"
@@ -393,7 +412,15 @@ impl<'a> FunctionSignature<'a> {
                         arg.span(),
                     )?;
                     if let Some((_, default)) = &arg.eq_and_default {
-                        fn_arg.default = Some(default.clone());
+                        if let FnArg::Regular(arg) = fn_arg {
+                            arg.default_value = Some(default.clone());
+                        } else {
+                            unreachable!(
+                                "`Python` and `CancelHandle` are already handled above and `*args`/`**kwargs` are \
+                                parsed and transformed below. Because the have to come last and are only allowed \
+                                once, this has to be a regular argument."
+                            );
+                        }
                     }
                 }
                 SignatureItem::VarargsSep(sep) => {
@@ -401,12 +428,12 @@ impl<'a> FunctionSignature<'a> {
                 }
                 SignatureItem::Varargs(varargs) => {
                     let fn_arg = next_non_py_argument_checked(&varargs.ident)?;
-                    fn_arg.is_varargs = true;
+                    fn_arg.to_varargs_mut()?;
                     parse_state.add_varargs(&mut python_signature, varargs)?;
                 }
                 SignatureItem::Kwargs(kwargs) => {
                     let fn_arg = next_non_py_argument_checked(&kwargs.ident)?;
-                    fn_arg.is_kwargs = true;
+                    fn_arg.to_kwargs_mut()?;
                     parse_state.add_kwargs(&mut python_signature, kwargs)?;
                 }
                 SignatureItem::PosargsSep(sep) => {
@@ -416,9 +443,11 @@ impl<'a> FunctionSignature<'a> {
         }
 
         // Ensure no non-py arguments remain
-        if let Some(arg) = args_iter.find(|arg| !arg.py) {
+        if let Some(arg) =
+            args_iter.find(|arg| !matches!(arg, FnArg::Py(..) | FnArg::CancelHandle(..)))
+        {
             bail_spanned!(
-                attribute.kw.span() => format!("missing signature entry for argument `{}`", arg.name)
+                attribute.kw.span() => format!("missing signature entry for argument `{}`", arg.name())
             );
         }
 
@@ -429,136 +458,27 @@ impl<'a> FunctionSignature<'a> {
         })
     }
 
-    /// The difference to `from_arguments_and_signature` is that deprecated args allowed entries to be:
-    ///  - missing
-    ///  - out of order
-    pub fn from_arguments_and_deprecated_args(
-        mut arguments: Vec<FnArg<'a>>,
-        deprecated_args: DeprecatedArgs,
-    ) -> syn::Result<Self> {
-        let mut varargs = None;
-        let mut kwargs = None;
-        let mut keyword_only_parameters = Vec::new();
-
-        fn first_n_argument_names(arguments: &[FnArg<'_>], count: usize) -> Vec<String> {
-            arguments
-                .iter()
-                .filter_map(|fn_arg| {
-                    if fn_arg.py {
-                        None
-                    } else {
-                        Some(fn_arg.name.unraw().to_string())
-                    }
-                })
-                .take(count)
-                .collect()
-        }
-
-        // Record highest counts observed based off argument positions
-        let mut positional_only_arguments_count = None;
-        let mut positional_arguments_count = None;
-        let mut required_positional_parameters = 0;
-
-        let args_iter = arguments.iter_mut().filter(|arg| !arg.py); // Python<'_> arguments don't show on the Python side.
-
-        for (i, fn_arg) in args_iter.enumerate() {
-            if let Some(argument) = deprecated_args
-                .arguments
-                .iter()
-                .find(|argument| match argument {
-                    Argument::PosOnlyArg(path, _)
-                    | Argument::Arg(path, _)
-                    | Argument::Kwarg(path, _)
-                    | Argument::VarArgs(path)
-                    | Argument::KeywordArgs(path) => path.get_ident() == Some(fn_arg.name),
-                    _ => false,
-                })
-            {
-                match argument {
-                    Argument::PosOnlyArg(_, default) | Argument::Arg(_, default) => {
-                        if let Some(default) = default {
-                            fn_arg.default = Some(syn::parse_str(default)?);
-                        } else if fn_arg.optional.is_none() {
-                            // Option<_> arguments always have an implicit None default with the old
-                            // `#[args]`
-                            required_positional_parameters = i + 1;
-                        }
-                        if matches!(argument, Argument::PosOnlyArg(_, _)) {
-                            positional_only_arguments_count = Some(i + 1);
-                        }
-                        positional_arguments_count = Some(i + 1);
-                    }
-                    Argument::Kwarg(_, default) => {
-                        fn_arg.default = default.as_deref().map(syn::parse_str).transpose()?;
-                        keyword_only_parameters.push((fn_arg.name.to_string(), default.is_none()));
-                    }
-                    Argument::PosOnlyArgsSeparator => {}
-                    Argument::VarArgsSeparator => {}
-                    Argument::VarArgs(path) => {
-                        fn_arg.is_varargs = true;
-                        if let Some(ident) = path.get_ident() {
-                            varargs = Some(ident.to_string());
-                        } else {
-                            bail_spanned!(path.span() => "expected ident for *args");
-                        };
-                    }
-                    Argument::KeywordArgs(path) => {
-                        fn_arg.is_kwargs = true;
-                        if let Some(ident) = path.get_ident() {
-                            kwargs = Some(ident.to_string());
-                        } else {
-                            bail_spanned!(path.span() => "expected ident for **kwargs");
-                        };
-                    }
-                }
-            } else {
-                // Assume this is a required positional parameter
-                required_positional_parameters = i + 1;
-                positional_arguments_count = Some(i + 1);
-            }
-        }
-
-        // fix up state based on observations above
-        let positional_only_parameters = positional_only_arguments_count.unwrap_or(0);
-        let positional_parameters = first_n_argument_names(
-            &arguments,
-            max(
-                positional_arguments_count.unwrap_or(0),
-                positional_only_arguments_count.unwrap_or(0),
-            ),
-        );
-
-        Ok(FunctionSignature {
-            arguments,
-            python_signature: PythonSignature {
-                positional_parameters,
-                positional_only_parameters,
-                required_positional_parameters,
-                varargs,
-                keyword_only_parameters,
-                kwargs,
-            },
-            attribute: None,
-        })
-    }
-
     /// Without `#[pyo3(signature)]` or `#[args]` - just take the Rust function arguments as positional.
-    pub fn from_arguments(mut arguments: Vec<FnArg<'a>>, deprecations: &mut Deprecations) -> Self {
+    pub fn from_arguments(arguments: Vec<FnArg<'a>>) -> syn::Result<Self> {
         let mut python_signature = PythonSignature::default();
         for arg in &arguments {
             // Python<'_> arguments don't show in Python signature
-            if arg.py {
+            if matches!(arg, FnArg::Py(..) | FnArg::CancelHandle(..)) {
                 continue;
             }
 
-            if arg.optional.is_none() {
-                // This argument is required
-                if python_signature.required_positional_parameters
-                    != python_signature.positional_parameters.len()
-                {
-                    // A previous argument was not required
-                    deprecations.push(Deprecation::RequiredArgumentAfterOption, arg.name.span());
-                }
+            if let FnArg::Regular(RegularArg {
+                ty,
+                option_wrapped_type: None,
+                ..
+            }) = arg
+            {
+                // This argument is required, all previous arguments must also have been required
+                ensure_spanned!(
+                    python_signature.required_positional_parameters == python_signature.positional_parameters.len(),
+                    ty.span() => "required arguments after an `Option<_>` argument are ambiguous\n\
+                    = help: add a `#[pyo3(signature)]` annotation on this function to unambiguously specify the default values for all optional parameters"
+                );
 
                 python_signature.required_positional_parameters =
                     python_signature.positional_parameters.len() + 1;
@@ -566,29 +486,24 @@ impl<'a> FunctionSignature<'a> {
 
             python_signature
                 .positional_parameters
-                .push(arg.name.unraw().to_string());
+                .push(arg.name().unraw().to_string());
         }
 
-        // Fixup any `Option<_>` arguments that were made implicitly made required by the deprecated
-        // branch above
-        for arg in arguments
-            .iter_mut()
-            .take(python_signature.required_positional_parameters)
-        {
-            arg.optional = None;
-        }
-
-        Self {
+        Ok(Self {
             arguments,
             python_signature,
             attribute: None,
-        }
+        })
     }
 
     fn default_value_for_parameter(&self, parameter: &str) -> String {
         let mut default = "...".to_string();
-        if let Some(fn_arg) = self.arguments.iter().find(|arg| arg.name == parameter) {
-            if let Some(arg_default) = fn_arg.default.as_ref() {
+        if let Some(fn_arg) = self.arguments.iter().find(|arg| arg.name() == parameter) {
+            if let FnArg::Regular(RegularArg {
+                default_value: Some(arg_default),
+                ..
+            }) = fn_arg
+            {
                 match arg_default {
                     // literal values
                     syn::Expr::Lit(syn::ExprLit { lit, .. }) => match lit {
@@ -614,7 +529,11 @@ impl<'a> FunctionSignature<'a> {
                     // others, unsupported yet so defaults to `...`
                     _ => {}
                 }
-            } else if fn_arg.optional.is_some() {
+            } else if let FnArg::Regular(RegularArg {
+                option_wrapped_type: Some(..),
+                ..
+            }) = fn_arg
+            {
                 // functions without a `#[pyo3(signature = (...))]` option
                 // will treat trailing `Option<T>` arguments as having a default of `None`
                 default = "None".to_string();
